local mymodule = {}

-- Load libraries
posix = require("posix")
modelfunctions = require("modelfunctions")
dbmodelfunctions = require("dbmodelfunctions")
fs = require("acf.fs")
format = require("acf.format")
validator = require("acf.validator")
db = require("acf.db")

-- Set variables
local processname = "kamailio"
local packagename = "kamailio"
local baseurl = "/etc/kamailio"
local kamctlrc_file = "/etc/kamailio/kamctlrc"

local dbkam

-- ################################################################################
-- DATABASE FUNCTIONS

local function assert (v, m)
	if not v then
		m = m or "Assertion failed!"
		error(m, 0)
	end
	return v, m
end

local databasecreate = function()
	if not dbkam then
		local engine
		-- parse the kamctlrc file
		local config = format.parse_ini_file(fs.read_file(kamctlrc_file), "") or {}
		if not config.DBENGINE then
			error("Database engine not specified, please setup one in the config script "..kamctlrc_file)
		end

		-- create environment object
		if config.DBENGINE == "MYSQL" or config.DBENGINE == "mysql" or config.DBENGINE == "MySQL" then
			error("MYSQL database not supported")
		elseif config.DBENGINE == "PGSQL" or config.DBENGINE == "pgsql" or config.DBENGINE == "postgres" or config.DBENGINE == "postgresql" or config.DBENGINE == "POSTGRESQL" then
			engine = db.engine.postgresql
		elseif config.DBENGINE == "ORACLE" or config.DBENGINE == "oracle" or config.DBENGINE == "Oracle" then
			error("ORACLE database not supported")
		elseif config.DBENGINE == "DBTEXT" or config.DBENGINE == "dbtext" or config.DBENGINE == "textdb" then
			error("DBTEXT database not supported")
		elseif config.DBENGINE == "DB_BERKELEY" or config.DBENGINE == "db_berkeley" or config.DBENGINE == "BERKELEY" or config.DBENGINE == "berkeley" then
			error("BERKELEY database not supported")
		else
			error("Unknown database engine "..config.DBENGINE)
		end

		-- connect to data source
		dbkam = db.create(engine, config.DBNAME or "openser", config.DBRWUSER or "openser", config.DBRWPW or "openserrw", config.DBHOST, config.DBPORT)
	end
end

local databaseconnect = function()
	if not dbkam then
		databasecreate()
	end
	return dbkam.databaseconnect()
end

local databasedisconnect = function()
	local result = dbkam.databasedisconnect()
	dbkam = nil
	return result
end

local runsqlcommand = function(sql)
	mymodule.logevent(sql)
	return dbkam.runsqlcommand(sql)
end

local getselectresponse = function(sql)
	mymodule.logevent(sql)
	return dbkam.getselectresponse(sql)
end

-- ################################################################################
-- LOCAL FUNCTIONS

local is_valid_filename = function(filename)
	local dirname = posix.dirname(filename)
	return validator.is_valid_filename(filename) and string.match(dirname, baseurl) and not string.match(dirname, "%.%.")
end

local function validate_user(user)
	local success = true
	if user.value.username.value == "" then
		user.value.username.errtxt = "Cannot be empty"
		success = false
	end
	if user.value.password.value == "" then
		user.value.password.errtxt = "Cannot be empty"
		success = false
	end
	if user.value.password.value ~= user.value.password_confirm.value then
		user.value.password_confirm.errtxt = "Must match password"
		success = false
	end
	return success, user
end

-- ################################################################################
-- PUBLIC FUNCTIONS

function mymodule.get_startstop(self, clientdata)
        return modelfunctions.get_startstop(processname)
end
	
function mymodule.startstop_service(self, startstop, action)
        return modelfunctions.startstop_service(startstop, action)
end

function mymodule.getstatus()
	return modelfunctions.getstatus(processname, packagename, "Kamailio Status")
end

function mymodule.get_filedetails(self, clientdata)
	return modelfunctions.getfiledetails(clientdata.filename, is_valid_filename)
end

function mymodule.update_filedetails(self, filedetails)
	return modelfunctions.setfiledetails(self, filedetails, is_valid_filename)
end

function mymodule.list_files()
	local retval = {}
	for file in fs.find(null, baseurl) do
		local details = fs.stat(file)
		if details.type == "regular" then
			details.filename = file
			table.insert(retval, details)
		end
	end
	table.sort(retval, function(a,b) return a.filename < b.filename end)
	return cfe({ type="structure", value=retval, label="List of Kamailio files" })
end

local function parse_db_show(table)
	-- These settings work for Postgres and DBTEXT database
	local delimiter = "\'?%s*[,|]%s*\'?"
	local results = {}
	local errtxt
	local f
	local res, err = pcall(function()
		local connected = databaseconnect()
		f = modelfunctions.run_executable({"kamctl", "db", "show", dbkam.escape(table)})
		if connected then databasedisconnect() end
	end)
	if not res and err then
		errtxt = err
	else
		for line in string.gmatch(f, "[^\n]+") do
			if #results == 0 and string.match(line, "^ERROR:") then
				errtxt = line
				results = nil
				break
			end
			if string.match(line, "^[+-]+$") then
				results = {}
			else
				local words = format.string_to_table(line, delimiter)
				if words and #words > 0 then
					results[#results+1] = words
				end
			end
		end
	end
	return results, errtxt
end

function mymodule.list_users()
	-- Database format: id | username | domain | password | email_address | ha1 | ha1b | rpid
	local results = {}
	local r, errtxt
	r, errtxt = parse_db_show("subscriber")
	for i,words in ipairs(r or {}) do
		if #words > 1 then
			local temp = {username = words[2],
				--domain = words[3],
				password = words[4],
				--email_address = words[5]
			}
			results[#results+1] = temp
		end
	end
	table.sort(results, function(a,b) return a.username < b.username end)
	return cfe({type="list", value=results, label="Kamailio Users", errtxt=errtxt})
end

function mymodule.get_new_user()
	local user = {}
	user.username = cfe({label="User Name", seq=1})
	user.password = cfe({type="password", label="Password", seq=2})
	user.password_confirm = cfe({type="password", label="Password (confirm)", seq=3})
	--user.email_address = cfe({label="E-mail Address", seq=4})
	return cfe({type="group", value=user, label="Kamailio User"})
end

function mymodule.create_new_user(self, user)
	local success = validate_user(user)
	if success then
		local res, err = pcall(function()
			local connected = databaseconnect()
			user.descr, user.errtxt = modelfunctions.run_executable({"kamctl", "add", dbkam.escape(user.value.username.value), dbkam.escape(user.value.password.value)})
			if connected then databasedisconnect() end
		end)
		if not res and err then
			user.errtxt = err
		end
	else
		user.errtxt = "Failed to create new user"
	end

	return user
end

function mymodule.get_delete_user(self, clientdata)
	local user = {}
	user.username = cfe({label="User Name"})
	return cfe({type="group", value=user, label="Delete User"})
end

function mymodule.delete_user(self, user)
	if user.value.username.value == "" then
		user.value.username.errtxt = "Invalid username"
		user.errtxt = "Failed to delete user"
	else
		local res, err = pcall(function()
			local connected = databaseconnect()
			user.descr, user.errtxt = modelfunctions.run_executable({"kamctl", "rm", dbkam.escape(user.value.username.value)})
			if connected then databasedisconnect() end
		end)
		if not res and err then
			user.errtxt = err
		end
	end
	return user
end

function mymodule.get_user(self, clientdata)
	local user = mymodule.get_new_user()
	user.value.username.value = clientdata.username or ""
	user.value.username.errtxt = "Invalid user"
	user.value.username.readonly = true
	local users = mymodule.list_users()
	for i,u in ipairs(users.value) do
		if u.username == clientdata.username then
			user.value.username.errtxt = nil
			break
		end
	end
	return user
end

function mymodule.update_user(self, user)
	local success = validate_user(user)
	if success then
		local res, err = pcall(function()
			local connected = databaseconnect()
			user.descr, user.errtxt = modelfunctions.run_executable({"kamctl", "passwd", dbkam.escape(user.value.username.value), dbkam.escape(user.value.password.value)})
			if connected then databasedisconnect() end
		end)
		if not res and err then
			user.errtxt = err
		end
	else
		user.errtxt = "Failed to update user"
	end

	return user
end

function mymodule.get_create_database(self, clientdata)
	return cfe({ type="group", value={}, label="Create Database" })
end

function mymodule.create_database(self, create_db)
	create_db.descr, create_db.errtxt = modelfunctions.run_executable({"kamdbctl", "create"}, true, 'y\ny\n')
	return create_db
end

function mymodule.get_search_database(self, clientdata)
	local errtxt
	retval = {}
	retval.id = cfe({type="select", value=clientdata.id or "", label="Table.Column", option={}, seq=1})
	retval.comparison = cfe({type="select", value=clientdata.comparison or "=", label="Comparison", option={"=", "!=", "~", "!~", "~*", "!*~"}, seq=2})
	retval.value = cfe({label="Value", value=clientdata.value or "", descr="Value or SQL regular expression", seq=3})
        local res, err = pcall(function()
		local connected = databaseconnect()
		local tables = dbkam.listtables() or {}
		for i,t in ipairs(tables) do
			local columns = dbkam.listcolumns(t) or {}
			for i,c in ipairs(columns) do
				retval.id.option[#retval.id.option + 1] = t.."."..c
			end
		end
		if connected then databasedisconnect() end
	end)
	if not res and err then
		errtxt = err
	end
	return cfe({type="group", value=retval, label="Database Search", errtxt=errtxt})
end

function mymodule.search_database(self, search)
	if modelfunctions.validateselect(search.value.id) and modelfunctions.validateselect(search.value.comparison) then
	        local res, err = pcall(function()
			local connected = databaseconnect()
			-- Get the rows from the DB
			search.value.result = cfe({type="structure", value={}, label="List of Rows", seq=4 })
			local table, column = string.match(search.value.id.value, "^([^.]*)%.(.*)")
			if table then
				local sql = "SELECT * FROM "..dbkam.escape(table).." WHERE "..dbkam.escape(column)..dbkam.escape(search.value.comparison.value).."'"..dbkam.escape(search.value.value.value).."'"
				search.value.result.value = getselectresponse(sql)
			end
			if connected then databasedisconnect() end
		end)
		if not res and err then
			search.errtxt = err
		end
	else
		search.errtxt = "Failed to search database"
	end
	return search
end

for n,f in pairs(dbmodelfunctions) do
	mymodule[n] = function(...)
		databasecreate()
		return f(dbkam, ...)
	end
end

return mymodule
